import { randomID } from '../../util/helpers.js';
import { Action } from '../display/action.js';
import { Actor } from '../display/actor.js';
import { PostContent } from '../supporting/post-content.js';

class Request {
  static nextSeq = 0;

  constructor(fee, content) {
    this.seq = this.nextSeq;
    this.nextSeq += 1;
    this.id = `req_${randomID()}`;
    this.fee = fee;
    this.content = content;
    this.worker = null;
  }
}

/**
 * Purpose: Enable fee-driven work requests, to be completed by workers from the availability pool
 */
export class Business extends Actor {
  constructor(dao, name, scene) {
    super(name, scene);
    this.dao = dao;

    this.actions = {
      assignWork: new Action('assign work', scene),
      submitPost: new Action('submit post', scene),
      initiateValidationPool: new Action('initiate validation pool', scene),
    };

    this.requests = new Map();
  }

  async submitRequest(fee, content) {
    const request = new Request(fee, content);
    this.requests.set(request.id, request);
    await this.actions.assignWork.log(this, this.dao.availability);
    const worker = await this.dao.availability.assignWork(request.id);
    request.worker = worker;
    return request.id;
  }

  async getRequest(requestId) {
    const request = this.requests.get(requestId);
    return request;
  }

  async getRequests() {
    return Array.from(this.requests.values());
  }

  async submitWork(reputationPublicKey, requestId, workEvidence, { tokenLossRatio, duration }) {
    const request = this.requests.get(requestId);
    if (!request) {
      throw new Error(`Request not found! id: ${requestId}`);
    }

    if (reputationPublicKey !== request.worker.reputationPublicKey) {
      throw new Error('Work evidence must be submitted by the assigned worker!');
    }

    // Create a post representing this submission.
    const post = new PostContent({
      requestId,
      workEvidence,
    });

    const requestIndex = Array.from(this.requests.values())
      .findIndex(({ id }) => id === request.id);

    post.setTitle(`Work Evidence ${requestIndex + 1}`);

    await this.actions.submitPost.log(this, this.dao);
    const { id: postId } = await this.dao.forum.addPost(reputationPublicKey, post);

    // Initiate a validation pool for this work evidence.
    await this.actions.initiateValidationPool.log(this, this.dao);
    const pool = await this.dao.initiateValidationPool(this, {
      postId,
      fee: request.fee,
      reputationPublicKey,
    }, {
      duration,
      tokenLossRatio,
    });

    await pool.stake(reputationPublicKey, {
      tokenAddress: request.worker.tokenAddress,
      amount: request.worker.stakeAmount,
      position: true,
    });

    // When the validation pool concludes,
    // reputation should be awarded and fees should be distributed.
    return pool;
  }
}
