import { Action } from '../display/action.js';
import { Actor } from '../display/actor.js';
import { ReputationHolder } from '../reputation/reputation-holder.js';
import { displayNumber } from '../../util/helpers.js';
import {
  EPSILON, INCINERATOR_ADDRESS, EdgeTypes, VertexTypes,
} from '../../util/constants.js';
import { WDGDiagram } from '../display/wdg-mermaid-ui.js';

class Post extends Actor {
  constructor(forum, senderId, postContent) {
    const index = forum.graph.countVertices(VertexTypes.POST);
    const name = `Post${index + 1}`;
    super(name, forum.scene);
    this.forum = forum;
    this.id = postContent.id ?? name;
    this.senderId = senderId;
    this.values = new Map();
    this.initialValues = new Map();
    this.authors = postContent.authors;
    this.citations = postContent.citations;
    this.title = postContent.title;
  }

  async setValue(value) {
    this.value = value;
    await this.setDisplayValue('value', value);
    this.forum.graph.getVertex(this.id).setProperty('value', value).displayVertex();
  }

  setInitialValue(value) {
    this.initialValue = value;
    this.forum.graph.getVertex(this.id).setProperty('initialValue', value).displayVertex();
  }
}

/**
 * Purpose:
 * - Forum: Maintain a directed, acyclic, graph of positively and negatively weighted citations.
 *   and the value accrued via each post and citation.
 */
export class Forum extends ReputationHolder {
  constructor(dao, name, scene) {
    super(name, scene);
    this.dao = dao;
    this.id = this.reputationPublicKey;
    this.graph = new WDGDiagram('forum', scene);
    this.actions = {
      propagate: new Action('propagate', scene),
      confirm: new Action('confirm', scene),
      transfer: new Action('transfer', scene),
    };
  }

  async addPost(senderId, postContent) {
    console.log('addPost', { senderId, postContent });
    const post = new Post(this, senderId, postContent);
    this.graph.addVertex(VertexTypes.POST, post.id, post, post.name);
    for (const { postId: citedPostId, weight } of post.citations) {
      // Special case: Incinerator
      if (citedPostId === INCINERATOR_ADDRESS && !this.graph.getVertex(INCINERATOR_ADDRESS)) {
        this.graph.addVertex(VertexTypes.POST, INCINERATOR_ADDRESS, { name: 'Incinerator' }, 'Incinerator');
      }
      this.graph.addEdge(EdgeTypes.CITATION, post.id, citedPostId, weight);
    }
    return post;
  }

  getPost(postId) {
    return this.graph.getVertexData(postId);
  }

  getPosts() {
    return this.graph.getVerticesData();
  }

  getTotalValue() {
    return this.getPosts().reduce((total, { value }) => total += value, 0);
  }

  // getLatestContract(type) { }

  // getContract(type) { }

  async onValidate({
    pool, postId, tokenAddress, referenceChainLimit, leachingValue,
  }) {
    console.log('onValidate', { pool, postId, tokenAddress });

    // What we have here now is an ERC-1155 rep token, which can contain multiple reputation types.
    // TODO: ERC-1155 supports a batch transfer operation, so it probably makes sense to leverage that.

    const initialValues = new Map(pool.reputationTypeIds
      .map((tokenTypeId) => [tokenTypeId, this.dao.reputation.valueOf(tokenAddress, tokenTypeId)]));
    const postVertex = this.graph.getVertex(postId);
    const post = postVertex.data;
    post.setStatus('Validated');
    post.initialValues = initialValues;

    const addAuthorToGraph = (publicKey, weight, authorTokenAddress) => {
      // For graph display purposes, we want to use the existing Expert actors from the current scene.
      const author = this.scene.findActor(({ reputationPublicKey }) => reputationPublicKey === publicKey);
      author.setDisplayValue('reputation', () => author.getReputation());
      const authorVertex = this.graph.getVertex(publicKey)
        ?? this.graph.addVertex(VertexTypes.AUTHOR, publicKey, author, author.name, {
          hide: author.options.hide,
        });
      this.graph.addEdge(
        EdgeTypes.AUTHOR,
        postVertex,
        authorVertex,
        weight,
        { tokenAddress: authorTokenAddress },
        { hide: author.options.hide },
      );
    };

    // In the case of multiple authors, mint additional (empty) tokens.
    // If no authors are specified, treat the sender as the sole author.
    // TODO: Verify that cumulative author weight == 1.
    if (!post.authors?.length) {
      addAuthorToGraph(post.senderId, 1, tokenAddress);
    } else {
      for (const { publicKey, weight } of post.authors) {
        // If the sender is also listed among the authors, do not mint them an additional token.
        const authorTokenAddress = (publicKey === post.senderId)
          ? tokenAddress
          : this.dao.reputation.mintBatch(this.id, pool.reputationTypeIds, pool.reputationTypeIds.map(() => 0));
        addAuthorToGraph(publicKey, weight, authorTokenAddress);
      }
      // If the sender is not an author, they will end up with the minted token but with zero value.
      if (!post.authors.find(({ publicKey }) => publicKey === post.senderId)) {
        addAuthorToGraph(post.senderId, 0, tokenAddress);
      }
    }

    const rewardsAccumulator = new Map();

    // Compute reputation rewards
    await this.propagateValue(
      { to: postVertex, from: { data: pool } },
      {
        rewardsAccumulator,
        increments: initialValues,
        referenceChainLimit,
        leachingValue,
      },
    );

    // Apply computed rewards to update values of tokens
    for (const [authorEdge, amount] of rewardsAccumulator) {
      const { to: authorVertex, data: { tokenAddress: authorTokenAddress } } = authorEdge;
      const { data: author } = authorVertex;
      // The primary author gets the validation pool minted token.
      // So we don't need to transfer any reputation to the primary author.
      // Their reward will be the remaining balance after all other transfers.
      if (authorTokenAddress !== tokenAddress) {
        if (amount < 0) {
          this.dao.reputation.transferValueFrom(authorTokenAddress, tokenAddress, -amount);
        } else {
          this.dao.reputation.transferValueFrom(tokenAddress, authorTokenAddress, amount);
        }
        await author.computeDisplayValues((label, value) => authorVertex.setProperty(label, value));
        authorVertex.displayVertex();
      }
    }

    const senderVertex = this.graph.getVertex(post.senderId);
    const { data: sender } = senderVertex;
    await sender.computeDisplayValues((label, value) => senderVertex.setProperty(label, value));
    senderVertex.displayVertex();

    // Transfer ownership of the minted tokens to the authors
    for (const authorEdge of postVertex.getEdges(EdgeTypes.AUTHOR, true)) {
      const authorVertex = authorEdge.to;
      const author = authorVertex.data;
      const { tokenAddress: authorTokenAddress } = authorEdge.data;
      this.dao.reputation.transfer(this.id, author.reputationPublicKey, authorTokenAddress);
    }
  }

  /**
   * @param {Edge} edge
   * @param {Object} opaqueData
   */
  async propagateValue(edge, {
    rewardsAccumulator,
    increments,
    depth = 0,
    initialNegative = false,
    referenceChainLimit,
    leachingValue,
  }) {
    const postVertex = edge.to;
    const post = postVertex.data;
    const incrementsStrArray = Array.from(increments.entries()).map(([k, v]) => `${k} ${v}`);
    const incrementsStr = `(${incrementsStrArray.join(')(')})`;
    this.actions.propagate.log(edge.from.data, post, incrementsStr);

    if (!!referenceChainLimit && depth > referenceChainLimit) {
      this.actions.propagate.log(
        edge.from.data,
        post,
        `referenceChainLimit (${referenceChainLimit}) reached`,
        null,
        '-x',
      );
      return increments;
    }

    console.log('propagateValue start', {
      from: edge.from.id ?? edge.from,
      to: edge.to.id,
      depth,
      values: post.values,
      increments,
      initialNegative,
    });

    const propagate = async (positive) => {
      const totalOutboundAmounts = increments.map(() => 0);
      const citationEdges = postVertex.getEdges(EdgeTypes.CITATION, true)
        .filter(({ weight }) => (positive ? weight > 0 : weight < 0));
      for (const citationEdge of citationEdges) {
        const { weight } = citationEdge;
        const outboundAmounts = increments.map((increment) => weight * increment);
        const refundsFromOutbound = increments.map(() => 0);
        for (let idx = 0; idx < outboundAmounts.length; idx++) {
          let outboundAmount = outboundAmounts[idx];
          if (Math.abs(outboundAmount) > EPSILON) {
            const balanceToOutbound = this.graph.getEdgeWeight(EdgeTypes.BALANCE, citationEdge.from, citationEdge.to)
            ?? 0;
            let refundFromOutbound = 0;

            // Special case: Incineration.
            if (citationEdge.to.id === INCINERATOR_ADDRESS) {
            // Only a positive amount may be incinerated! Otherwise the sink could be used as a source.
              if (outboundAmount < 0) {
                this.scene?.flowchart?.log(`style ${citationEdge.from.id} fill:#620000`);
                this.actions.propagate.log(
                  citationEdge.from.data,
                  { name: 'Incinerator' },
                  incrementsStr,
                  undefined,
                  '-x',
                );
                throw new Error('Incinerator can only receive positive citations!');
              }
              // Reputation sent to the incinerator is burned! This means it is deducted from the sender,
              // without increasing the value of any other token.
              this.actions.propagate.log(citationEdge.from.data, { name: 'Incinerator' }, incrementsStr);
            } else {
            // We need to ensure that we at most undo the prior effects of this post
              if (initialNegative) {
                outboundAmount = outboundAmount < 0
                  ? Math.max(outboundAmount, -balanceToOutbound)
                  : Math.min(outboundAmount, -balanceToOutbound);
              }

              // Recursively propagate reputation effects
              refundFromOutbound = await this.propagateValue(citationEdge, {
                rewardsAccumulator,
                increment: outboundAmount,
                depth: depth + 1,
                initialNegative: initialNegative || (depth === 0 && outboundAmount < 0),
                referenceChainLimit,
                leachingValue,
              });

              // Any excess (negative) amount that could not be propagated,
              // i.e. because a cited post has been reduced to zero value,
              // is retained by the citing post.
              outboundAmount -= refundFromOutbound;
            }

            // Keep a record of the effect of the reputation transferred along this edge in the graph,
            // so that later, negative citations can be constrained to at most undo these effects.
            this.graph.setEdgeWeight(
              EdgeTypes.BALANCE,
              citationEdge.from,
              citationEdge.to,
              balanceToOutbound + outboundAmount,
            );
            refundsFromOutbound[idx] = refundFromOutbound;
            totalOutboundAmounts[idx] += outboundAmount;
          }
          const refundStr = refundsFromOutbound.map((refund) => displayNumber(refund)).join('/');
          this.actions.confirm.log(
            citationEdge.to.data,
            citationEdge.from.data,
            `(refund: ${refundStr}, leach: ${outboundAmount * leachingValue})`,
            undefined,
            '-->>',
          );
        }
      }
      return totalOutboundAmounts;
    };

    // First, leach value via negative citations
    const totalLeachingAmounts = await propagate(false);
    for (let idx = 0; idx < totalLeachingAmounts.length; idx++) {
      increments[idx] -= totalLeachingAmounts[idx] * leachingValue;
    }

    // Now propagate value via positive citations
    const totalDonationAmounts = await propagate(true);
    for (let idx = 0; idx < totalDonationAmounts.length; idx++) {
      increments[idx] -= totalDonationAmounts[idx] * leachingValue;
    }

    // Apply the remaining increment to the present post
    const rawNewValues = post.values.map((value, idx) => value + increments[idx]);
    const newValues = rawNewValues.map((rawNewValue) => Math.max(0, rawNewValue));
    const appliedIncrements = newValue - post.value;
    const refundToInbound = increment - appliedIncrement;

    // Apply reputation effects to post authors, not to the post directly
    for (const authorEdge of postVertex.getEdges(EdgeTypes.AUTHOR, true)) {
      const { weight, to: { data: author } } = authorEdge;
      const authorIncrement = weight * appliedIncrement;
      rewardsAccumulator.set(authorEdge, authorIncrement);
      this.actions.propagate.log(post, author, `(${authorIncrement})`);
    }

    console.log('propagateValue end', {
      depth,
      increment,
      rawNewValue,
      newValue,
      appliedIncrement,
      refundToInbound,
    });

    // Increment the value of the post
    await post.setValue(newValue);

    return refundToInbound;
  }
}
