import {
  DEFAULT_TIME_STEP,
  MINIMUM_VELOCITY,
  VISCOSITY_FACTOR,
  TIME_DILATION_FACTOR,
  MAX_STEPS_TO_EQUILIBRIUM,
  TRANSLATION_VELOCITY_FACTOR,
} from '../../util/constants.js';
import { Edge } from '../supporting/edge.js';
import { WeightedDirectedGraph } from '../supporting/wdg.js';
import { Box } from './box.js';
import { Vector, Rectangle } from '../supporting/geometry/index.js';
import { overlapRepulsionForce, targetRadiusForce } from './pairwise-forces.js';

// Render children with absolute css positioning.

// Let there be a force between elements such that the force between
// any two elements is along the line between their centers,
// so that the elements repel when too close but attract when too far.

// The equilibrium distance can be tuned, e.g. can be scaled by an input.

// NOTE: (with optional overlay preferring a grid or some other shape?),

// NOTE: Could also allow user input dragging elements.
// What might be neat here is to implement a force-based resistance effect;
// basically, the mouse pointer drags the element with a spring rather than directly.
// If the shape of the graph resists the transformation,
// the distance between the element and the cursor should increase.

// On an interval, compute forces among the elements.
// Simulate the effects of these forces

// NOTE: Impart random nudges, and resolve their effects to a user-visible resolution
// before rendering.

// NOTE: When mouse is in our box, we could hijack the scroll actions to zoom in/out.

export class ForceDirectedGraph extends WeightedDirectedGraph {
  constructor(name, parentEl, { width = 800, height = 600, ...options } = {}) {
    super(name, options);
    this.box = new Box(name, parentEl, options);
    this.box.addClass('fixed');
    this.box.addClass('force-directed-graph');
    this.intervalTask = null;
    this.canvas = window.document.createElement('canvas');
    this.width = width;
    this.height = height;
    this.box.el.style.width = `${width}px`;
    this.box.el.style.height = `${height}px`;
    this.box.el.appendChild(this.canvas);
    this.fitCanvasToGraph();
    this.nodes = [];
    this.edges = [];
  }

  fitCanvasToGraph() {
    [this.canvas.width, this.canvas.height] = this.box.rect.dimensions;
  }

  // addVertex(type, id, data, label, options) {
  addVertex(...args) {
    const vertex = super.addVertex(...args);
    const box = this.box.addBox(vertex.id);

    // Link from the graph vertex to the corresponding display box
    vertex.box = box;

    // Link from the display box to the corresponding graph vertex
    box.vertex = vertex;

    box.addClass('absolute');
    box.addClass('vertex');
    box.el.style.left = '0px';
    box.el.style.top = '0px';
    box.velocity = Vector.from([0, 0]);
    box.setInnerHTML(vertex.getHTML());
    this.nodes.push(box);

    // When vertex properties are updated, re-render the node contents
    vertex.onUpdate = () => {
      box.setInnerHTML(vertex.getHTML());
      // Maybe resolve forces
      this.runUntilEquilibrium();
    };

    this.runUntilEquilibrium();

    // Allow moving vertices with the mouse
    box.el.addEventListener('mousedown', (e) => {
      if (!this.mouseMoving) {
        e.preventDefault();
        // Record current mouse position
        this.mousePosition = Vector.from([e.clientX, e.clientY]);
        // Begin tracking mouse movement
        this.mouseMoving = box;
      }
    });
    document.addEventListener('mousemove', (e) => {
      if (this.mouseMoving === box) {
        const mousePosition = Vector.from([e.clientX, e.clientY]);
        // Apply translation
        box.move(mousePosition.subtract(this.mousePosition));
        // Update current mouse position
        this.mousePosition = mousePosition;
        // Equilibrate
        this.runUntilEquilibrium();
      }
    });
    document.addEventListener('mouseup', () => {
      // Stop tracking mouse movement
      this.mouseMoving = null;
      // Equilibrate
      this.runUntilEquilibrium();
    });

    return vertex;
  }

  addEdge(type, from, to, ...rest) {
    const fromBox = this.nodes.find(({ name }) => name === from);
    const toBox = this.nodes.find(({ name }) => name === to);
    if (!fromBox) throw new Error(`from ${from}: Node not found`);
    if (!toBox) throw new Error(`to ${to}: Node not found`);

    const edge = super.addEdge(type, from, to, ...rest);
    const box = this.box.addBox(Edge.getKey({ from, to, type }));

    edge.box = box;
    box.edge = edge;

    box.addClass('absolute');
    box.addClass('edge');

    box.setInnerHTML(edge.getHTML());
    this.edges.push(box);

    // Center between nodes, by applying an attraction force from the edge node
    // to its `from` and `to` nodes.
    // We can also initially place it near the midpoint between the `from` and `to` nodes.
    const midpoint = fromBox.rect.center.add(toBox.rect.center).scale(0.5);
    const startPosition = midpoint.subtract(box.rect.dimensions.scale(0.5));
    console.log({
      fromBox, toBox, midpoint, startPosition, dimensions: box.rect.dimensions,
    });
    box.move(startPosition);
    box.velocity = Vector.from([0, 0]);
    this.runUntilEquilibrium();

    return edge;
  }

  setEdgeWeight(...args) {
    const edge = super.setEdgeWeight(...args);
    edge.displayEdgeNode();
    return edge;
  }

  async runUntilEquilibrium(tDelta = DEFAULT_TIME_STEP) {
    this.steps = 0;
    if (this.intervalTask) {
      return Promise.resolve();
    }
    return new Promise((resolve, reject) => {
      this.intervalTask = setInterval(() => {
        this.steps++;
        if (this.steps > MAX_STEPS_TO_EQUILIBRIUM) {
          clearInterval(this.intervalTask);
          this.intervalTask = null;
          reject(new Error('Exceeded map steps to reach equilibrium'));
        }
        const { atEquilibrium } = this.computeEulerFrame(tDelta);
        if (atEquilibrium) {
          console.log(`Reached equilibrium after ${this.steps} steps`);
          clearInterval(this.intervalTask);
          this.intervalTask = null;
          resolve();
        }
      }, tDelta * TIME_DILATION_FACTOR);
    });
  }

  computeEulerFrame(tDelta = DEFAULT_TIME_STEP) {
    // Compute net forces on each box in the graph
    const boxes = [...this.nodes, ...this.edges];

    // Initialize net force vectors
    for (const box of boxes) {
      box.netForce = Vector.zeros(2);
    }

    // Compute overlap repulsion forces among node boxes
    for (const boxA of this.nodes) {
      const idxA = boxes.indexOf(boxA);
      for (const boxB of this.nodes.slice(idxA + 1)) {
        const force = overlapRepulsionForce(boxA, boxB);
        boxA.netForce = boxA.netForce.subtract(force);
        boxB.netForce = boxB.netForce.add(force);
      }
    }

    // Compute pairwise forces among nodes
    for (const boxA of this.nodes) {
      const idxA = this.nodes.indexOf(boxA);
      for (const boxB of this.nodes.slice(idxA + 1)) {
        const force = targetRadiusForce(boxA, boxB);
        boxA.netForce = boxA.netForce.subtract(force);
        boxB.netForce = boxB.netForce.add(force);
      }
    }

    // Compute forces on edge boxes:
    // Attraction to the `from` and `to` nodes
    for (const edgeBox of this.edges) {
      const { edge } = edgeBox;
      const fromBox = edge.from.box;
      const toBox = edge.to.box;
      edgeBox.netForce = edgeBox.netForce
        .subtract(targetRadiusForce(edgeBox, fromBox, 0))
        .subtract(targetRadiusForce(edgeBox, toBox, 0));
    }

    // Do not apply forces to a box if it is being moved by the mouse
    for (const box of boxes) {
      if (this.mouseMoving === box) {
        box.netForce = Vector.zeros(2);
      }
    }

    // Compute velocities
    for (const box of boxes) {
      box.velocity = box.velocity.add(box.netForce.scale(tDelta));
      // Apply some drag
      box.velocity = box.velocity.scale(1 - VISCOSITY_FACTOR);
    }

    // When all velocities are below MINIMUM_VELOCITY, we have reached equilibrium.
    let atEquilibrium = true;

    // Apply velocities
    for (const box of boxes) {
      if (box.velocity.magnitude >= MINIMUM_VELOCITY) {
        atEquilibrium = false;
        box.move(box.velocity);
      }
    }

    // Center the items by computing the bounding box and centering that
    if (!this.mouseMoving) {
      const topLeft = Vector.from(boxes[0].position);
      const bottomRight = Vector.from(boxes[0].position);
      for (const box of boxes) {
        for (const vertex of box.rect.vertices) {
          topLeft[0] = Math.min(topLeft[0], vertex[0]);
          topLeft[1] = Math.min(topLeft[1], vertex[1]);

          bottomRight[0] = Math.max(bottomRight[0], vertex[0]);
          bottomRight[1] = Math.max(bottomRight[1], vertex[1]);
        }
      }
      const boundingBox = new Rectangle(topLeft, bottomRight.subtract(topLeft));
      const graphCenter = Vector.from([this.width, this.height]).scale(0.5);
      const offset = graphCenter.subtract(boundingBox.center);
      const translate = offset.scale(TRANSLATION_VELOCITY_FACTOR);

      if (translate.magnitude >= MINIMUM_VELOCITY) {
        atEquilibrium = false;

        // Apply translations
        for (const box of boxes) {
          box.move(translate);
        }
      }
    }

    // TODO: Scaling to fit

    return { atEquilibrium };
  }
}
