import { Vertex } from '../supporting/vertex.js';
import { Edge } from '../supporting/edge.js';
import { Document } from './document.js';
import { WeightedDirectedGraph } from '../supporting/wdg.js';

const allGraphs = [];

const makeWDGHandler = (graphIndex) => (vertexId) => {
  const graph = allGraphs[graphIndex];
  // We want a document for editing this node, which may be a vertex or an edge
  const { editorDoc } = graph;
  editorDoc.clear();
  if (vertexId.startsWith('edge:')) {
    const [, from, to] = vertexId.split(':');
    Edge.prepareEditorDocument(graph, editorDoc, from, to);
  } else {
    Vertex.prepareEditorDocument(graph, editorDoc, vertexId);
  }
};

export class WDGDiagram extends WeightedDirectedGraph {
  constructor(name, scene, options = {}) {
    super(name, options);
    this.scene = scene;
    this.flowchart = scene?.flowchart;
    this.editable = options.editable;

    // Mermaid supports a click callback, but we can't customize arguments; we just get the vertex ID.
    // In order to provide the appropriate graph context for each callback, we create a separate callback
    // function for each graph.
    this.index = allGraphs.length;
    allGraphs.push(this);
    window[`WDGHandler${this.index}`] = makeWDGHandler(this.index);
  }

  fromJSON(...args) {
    super.fromJSON(...args);
    this.redraw();
    return this;
  }

  addVertex(...args) {
    const vertex = super.addVertex(...args);
    vertex.displayVertex();
    return vertex;
  }

  addEdge(type, from, to, ...rest) {
    const existingEdges = this.getEdges(null, from, to);
    const edge = super.addEdge(type, from, to, ...rest);
    if (existingEdges.length) {
      edge.displayEdgeNode();
    } else {
      edge.displayEdge();
    }
    return edge;
  }

  setEdgeWeight(...args) {
    const edge = super.setEdgeWeight(...args);
    edge.displayEdgeNode();
    return edge;
  }

  redraw() {
    // Call .reset() on all vertices and edges
    for (const vertex of this.vertices.values()) {
      vertex.reset();
    }
    for (const edges of this.edgeTypes.values()) {
      for (const edge of edges.values()) {
        edge.reset();
      }
    }

    // Clear the target div
    this.flowchart?.reset();
    this.flowchart?.init();

    // Draw all vertices and edges
    for (const vertex of this.vertices.values()) {
      vertex.displayVertex();
    }
    // Let's flatmap and dedupe by [from, to] since each edge
    // renders all comorphic edges as well.
    const edgesFrom = new Map(); // edgeMap[from][to] = edge
    for (const edges of this.edgeTypes.values()) {
      for (const edge of edges.values()) {
        const edgesTo = edgesFrom.get(edge.from) || new Map();
        edgesTo.set(edge.to, edge);
        edgesFrom.set(edge.from, edgesTo);
      }
    }

    for (const edgesTo of edgesFrom.values()) {
      for (const edge of edgesTo.values()) {
        edge.displayEdge();
      }
    }

    // Ensure rerender
    this.flowchart?.render();
  }

  withFlowchart() {
    this.scene?.withSectionFlowchart();
    this.flowchart = this.scene?.lastFlowchart;
    if (this.editable) {
      this.controlDoc = new Document('WDGControl', this.flowchart.box.el, { prepend: true });
      this.editorDoc = new Document('WDGEditor', this.flowchart.box.el);
      this.errorDoc = new Document('WDGErrors', this.flowchart.box.el);
      this.resetEditor();
    }
    return this;
  }

  prepareControlDoc() {
    const form = this.controlDoc.form({ name: 'controlForm' }).lastElement;
    const { subForm: graphPropertiesForm } = form.subForm({ name: 'graphPropsForm' }).lastItem;
    graphPropertiesForm.flex()
      .textField({
        name: 'name',
        label: 'Graph name',
        defaultValue: this.name,
        cb: (({ form: { value: { name } } }) => {
          this.name = name;
        }),
      });
    const { subForm: exportImportForm } = form.subForm({ name: 'exportImportForm' }).lastItem;
    exportImportForm.flex()
      .button({
        name: 'Export',
        cb: () => {
          const a = window.document.createElement('a');
          const json = JSON.stringify(this.toJSON(), null, 2);
          const currentTime = Math.floor(new Date().getTime() / 1000);
          a.href = `data:attachment/text,${encodeURI(json)}`;
          a.target = '_blank';
          a.download = `wdg_${this.name}_${currentTime}.json`;
          a.click();
        },
      })
      .fileInput({
        name: 'Import',
        cb: ({ input: { files: [file] } }) => {
          const reader = new FileReader();
          reader.onload = ({ target: { result: text } }) => {
            console.log('imported file', { file });
            // this.flowchart?.log(`%% Imported file ${file}`)
            const data = JSON.parse(text);
            this.fromJSON(data);
          };
          reader.readAsText(file);
        },
      });
  }

  resetEditor() {
    this.editorDoc.clear();
    this.controlDoc.clear();
    this.prepareControlDoc();
    Vertex.prepareEditorDocument(this, this.editorDoc);
  }
}
