/**
 * ERC-1155: Multi Token Standard
 * See https://eips.ethereum.org/EIPS/eip-1155
 * and https://github.com/OpenZeppelin/openzeppelin-contracts/blob/master/contracts/token/ERC1155/ERC1155.sol
 *
 * This implementation is currently incomplete. It lacks the following:
 * - Token approvals
 * - Operator approvals
 * - Emitting events
 * - transferFrom
 */

export class ERC1155 {
  constructor(name, symbol) {
    this.name = name;
    this.symbol = symbol;
    this.balances = new Map(); // owner address --> token id --> token count
    this.owners = new Map(); // token address --> owner address
    // this.tokenApprovals = new Map(); // token address --> approved addresses
    // this.operatorApprovals = new Map(); // ownerAddress --> operator approvals

    this.events = {
      // Transfer: (_from, _to, _tokenAddress) => {},
      // Approval: (_owner, _approved, _tokenAddress) => {},
      // ApprovalForAll: (_owner, _operator, _approved) => {},
    };
  }

  incrementBalance(ownerAddress, tokenTypeId, increment) {
    const tokens = this.balances.get(ownerAddress) ?? new Map();
    const balance = tokens.get(tokenTypeId) ?? 0;
    tokens.set(tokenTypeId, balance + increment);
    this.balances.set(ownerAddress, tokens);
  }

  mintBatch(to, tokenAddress, tokenTypeIds, amounts = null) {
    if (!amounts) {
      amounts = tokenTypeIds.map(() => 1);
    }
    console.log('ERC1155.mintBatch', {
      to, tokenAddress, tokenTypeIds, amounts,
    });
    if (this.owners.get(tokenAddress)) {
      throw new Error('ERC1155: token already minted');
    }
    for (let idx = 0; idx < tokenTypeIds.length; idx++) {
      const tokenTypeId = tokenTypeIds[idx];
      const amount = amounts[idx];
      this.incrementBalance(to, tokenTypeId, amount);
    }
    this.owners.set(tokenAddress, to);
  }

  burn(tokenAddress, tokenTypeId, amount = 1) {
    const ownerAddress = this.owners.get(tokenAddress);
    this.incrementBalance(ownerAddress, tokenTypeId, -amount);
  }

  balanceOf(ownerAddress, tokenTypeId) {
    if (!ownerAddress) {
      throw new Error('ERC1155: address zero is not a valid owner');
    }
    const tokens = this.balances.get(ownerAddress) ?? new Map();
    return tokens.get(tokenTypeId) ?? 0;
  }

  ownerOf(tokenAddress) {
    const ownerAddress = this.owners.get(tokenAddress);
    if (!ownerAddress) {
      throw new Error(`ERC1155: invalid token address: ${tokenAddress}`);
    }
    return ownerAddress;
  }

  transfer(from, to, tokenAddress) {
    console.log('ERC1155.transfer', { from, to, tokenAddress });
    const ownerAddress = this.owners.get(tokenAddress);
    if (ownerAddress !== from) {
      throw new Error(`ERC1155: transfer from incorrect owner ${from}; should be ${ownerAddress}`);
    }
    this.incrementBalance(from, -1);
    this.incrementBalance(to, 1);
    this.owners.set(tokenAddress, to);
  }

  /// @notice Enable or disable approval for a third party ("operator") to manage
  ///  all of `msg.sender`'s assets
  /// @dev Emits the ApprovalForAll event. The contract MUST allow
  ///  multiple operators per ownerAddress.
  /// @param _operator Address to add to the set of authorized operators
  /// @param _approved True if the operator is approved, false to revoke approval
  // setApprovalForAll(_operator, _approved) {}

  /// @notice Get the approved address for a single NFT
  /// @dev Throws if `_tokenAddress` is not a valid NFT.
  /// @param _tokenAddress The NFT to find the approved address for
  /// @return The approved address for this NFT, or the zero address if there is none
  // getApproved(_tokenAddress) {}

  /// @notice Query if an address is an authorized operator for another address
  /// @param _owner The address that owns the NFTs
  /// @param _operator The address that acts on behalf of the ownerAddress
  /// @return True if `_operator` is an approved operator for `_owner`, false otherwise
  // isApprovedForAll(_owner, _operator) {}
}
