import { Document } from '../display/document.js';

export const PropertyTypes = {
  string: 'string',
  number: 'number',
};

class Property {
  constructor(name, type) {
    this.name = name;
    this.type = type;
  }
}

class NodeType {
  constructor(name) {
    this.name = name;
    this.properties = new Set();
  }

  addProperty(name, type) {
    this.properties.add(new Property(name, type));
  }
}

class EdgeType {
  constructor(name, fromNodeTypes, toNodeTypes) {
    this.name = name;
    this.fromNodeTypes = fromNodeTypes;
    this.toNodeTypes = toNodeTypes;
  }
}

export class Schema {
  constructor() {
    this.nodeTypes = new Set();
    this.edgeTypes = new Set();
  }

  addNodeType(name) {
    const nodeType = new NodeType(name);
    this.nodeTypes.add(nodeType);
  }

  addEdgeType(name) {
    const edgeType = new EdgeType(name);
    this.nodeTypes.add(edgeType);
  }
}

export class SchemaEditor {
  constructor(name, scene, options = {}) {
    this.name = name;
    this.scene = scene;
    this.options = options;

    this.schemaEditorBox = scene.middleSection.addBox('Schema Editor').flex();
    this.nodeEditorDoc = new Document('NodeSchemaEditor', this.schemaEditorBox.el);
    this.edgeEditorDoc = new Document('EdgeSchemaEditor', this.schemaEditorBox.el);
    this.prepareNodeEditorDocument();
    this.prepareEdgeEditorDocument();
  }

  prepareNodeEditorDocument(schema = new Schema()) {
    const doc = this.nodeEditorDoc;
    doc.clear();
    doc.remark('<h2>Node Types</h2>');
    const form = doc.form({ name: 'Node Types Editor' }).lastElement;
    const nodeTypesSubFormArray = form.subFormArray({ name: 'nodeTypes' }).lastItem;

    const addNodeForm = (name, properties) => {
      const nodeTypeForm = nodeTypesSubFormArray.subForm().lastSubForm;
      if (name) {
        nodeTypeForm.remark(`<h3>Node Type: ${name}</h3>`);
      } else {
        nodeTypeForm.remark('<h3>New Node Type</h3>');
      }
      nodeTypeForm.textField({ name: 'name', defaultValue: name });
      const propertiesSubFormArray = nodeTypeForm.subFormArray({ name: 'properties' }).lastItem;
      for (const property of properties.values()) {
        const propertyForm = propertiesSubFormArray.subForm().lastSubForm;
        propertyForm.textField({ name: 'name', defaultValue: property.name });
        propertyForm.textField({ name: 'type', defaultValue: property.type });
      }
    };

    for (const { name, properties } of schema.nodeTypes.values()) {
      addNodeForm(name, properties);
    }

    form.button({
      name: 'Add Node Type',
      cb: () => {
        addNodeForm('', new Set());
      },
    });

    form.submit({
      name: 'Save',
      cb: ({ form: { value: formValue } }) => {
        console.log('save', { formValue });
      },
    });
  }

  prepareEdgeEditorDocument(schema = new Schema()) {
    const doc = this.edgeEditorDoc;
    doc.clear();
    doc.remark('<h2>Edge Types</h2>');
    const form = doc.form('Edge Types Editor').lastElement;
    for (const { name } of schema.edgeTypes.values()) {
      form.remark(`<h3>Edge Type: ${name}</h3>`);
      form.textField({ name: 'name', defaultValue: name });
    }
    form.submit({
      name: 'Save',
    });
  }
}

// Properties
// Data types
// Relationships
