import { Vertex } from './vertex.js';
import { Edge } from './edge.js';

export class WeightedDirectedGraph {
  constructor(name) {
    this.name = name;
    this.vertices = new Map();
    this.edgeTypes = new Map();
    this.nextVertexId = 0;

    // TODO: Populate history
    this.history = [];
  }

  getHistory() {
    // record operations that modify the graph
    return this.history;
  }

  toJSON() {
    return {
      name: this.name,
      vertices: Array.from(this.vertices.values())
        .map((vertex) => vertex.toJSON()),
      edges: Array.from(this.edgeTypes.values())
        .flatMap((edges) => Array.from(edges.values()))
        .map((edge) => edge.toJSON()),
      // history: this.getHistory(),
    };
  }

  fromJSON({ name, vertices, edges }) {
    this.name = name;
    this.vertices.clear();
    this.edgeTypes.clear();
    for (const {
      id, type, label, properties,
    } of vertices) {
      const vertex = this.addVertex(type, id, null, label);
      for (const [key, value] of Object.entries(properties)) {
        vertex.setProperty(key, value);
      }
    }
    for (const {
      from, to, type, weight,
    } of edges) {
      this.addEdge(type, from, to, weight);
    }
    return this;
  }

  addVertex(type, id, data, label, options) {
    // Supports simple case of auto-incremented numeric ids
    if (typeof id === 'object') {
      data = id;
      id = this.nextVertexId++;
    }
    id = (typeof id === 'number') ? id.toString(10) : id;
    if (this.vertices.has(id)) {
      throw new Error(`Vertex already exists with id: ${id}`);
    }
    const vertex = new Vertex(this, type, id, data, { ...options, label });
    this.vertices.set(id, vertex);
    return vertex;
  }

  getVertex(id) {
    id = (typeof id === 'number') ? id.toString(10) : id;
    return this.vertices.get(id);
  }

  getVertexData(id) {
    return this.getVertex(id)?.data;
  }

  getVerticesData() {
    return Array.from(this.vertices.values()).map(({ data }) => data);
  }

  getEdge(type, from, to) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    if (!from || !to) {
      return undefined;
    }
    const edges = this.edgeTypes.get(type);
    const edgeKey = Edge.getKey({ from, to, type });
    return edges?.get(edgeKey);
  }

  getEdgeWeight(type, from, to) {
    return this.getEdge(type, from, to)?.weight;
  }

  setEdgeWeight(type, from, to, weight, data, options) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    const edge = new Edge(this, type, from, to, weight, data, options);
    let edges = this.edgeTypes.get(type);
    if (!edges) {
      edges = new Map();
      this.edgeTypes.set(type, edges);
    }
    const edgeKey = Edge.getKey(edge);
    edges.set(edgeKey, edge);
    return edge;
  }

  addEdge(type, from, to, weight, data, options) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    if (this.getEdge(type, from, to)) {
      throw new Error(`Edge ${type} from ${from.id} to ${to.id} already exists`);
    }
    const edge = this.setEdgeWeight(type, from, to, weight, data, options);
    from.edges.from.push(edge);
    to.edges.to.push(edge);
    return edge;
  }

  getEdges(type, from, to) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    const edgeTypes = type ? [type] : Array.from(this.edgeTypes.keys());
    return edgeTypes.flatMap((edgeType) => {
      const edges = this.edgeTypes.get(edgeType);
      return Array.from(edges?.values() || []).filter((edge) => {
        const matchFrom = from === null || from === undefined || from === edge.from;
        const matchTo = to === null || to === undefined || to === edge.to;
        return matchFrom && matchTo;
      });
    });
  }

  countVertices(type) {
    if (!type) {
      return this.vertices.size;
    }
    return Array.from(this.vertices.values()).filter((vertex) => vertex.type === type).length;
  }

  deleteEdge(type, from, to) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    const edges = this.edgeTypes.get(type);
    const edgeKey = Edge.getKey({ type, from, to });
    if (!edges) return;
    const edge = edges.get(edgeKey);
    if (!edge) return;
    to.edges.from.forEach((x, i) => (x === edge) && to.edges.from.splice(i, 1));
    from.edges.to.forEach((x, i) => (x === edge) && from.edges.to.splice(i, 1));
    edges.delete(edgeKey);
    if (edges.size === 0) {
      this.edgeTypes.delete(type);
    }
  }

  deleteVertex(id) {
    const vertex = this.getVertex(id);
    for (const { type, from, to } of [...vertex.edges.to, ...vertex.edges.from]) {
      this.deleteEdge(type, from, to);
    }
    this.vertices.delete(id);
  }
}
