import { Box } from '../../classes/display/box.js';
import { Scene } from '../../classes/display/scene.js';
import { WDGDiagram } from '../../classes/display/wdg-mermaid-ui.js';
import { WeightedDirectedGraph } from '../../classes/supporting/wdg.js';
import { mochaRun } from '../../util/helpers.js';

const rootElement = document.getElementById('scene');
const rootBox = new Box('rootBox', rootElement).flex();
window.scene = new Scene('WDG test', rootBox);

describe('Weighted Directed Graph', function tests() {
  this.timeout(0);

  let graph;

  before(() => {
    graph = (window.graph = new WDGDiagram('test1', window.scene)).withFlowchart();

    graph.addVertex('v1', {});
    graph.addVertex('v1', {});
    graph.addVertex('v1', {});
    graph.addVertex('v1', {});
    graph.addVertex('v1', {});

    graph.addEdge('e1', 0, 1, 1);
    graph.addEdge('e1', 2, 1, 0.5);
    graph.addEdge('e1', 3, 1, 0.25);
    graph.addEdge('e1', 1, 4, 0.125);
  });

  it('can query for all e1 edges', () => {
    const edges = graph.getEdges('e1');
    edges.should.have.length(4);
  });

  it('can query for all e1 edges from a particular vertex', () => {
    const edges = graph.getEdges('e1', 2);
    edges.map(({ from, to, weight }) => [from.id, to.id, weight]).should.have.deep.members([[
      '2', '1', 0.5,
    ]]);
  });

  it('can query for all e1 edges to a particular vertex', () => {
    const edges = graph.getEdges('e1', null, 1);
    edges.map(({ from, to, weight }) => [from.id, to.id, weight]).should.have.deep.members([
      ['0', '1', 1],
      ['2', '1', 0.5],
      ['3', '1', 0.25],
    ]);
  });

  it('can export to JSON', () => {
    const result = graph.toJSON();
    result.should.eql({
      name: 'test1',
      vertices: [
        {
          id: '0', type: 'v1', label: '0', properties: {},
        },
        {
          id: '1', type: 'v1', label: '1', properties: {},
        },
        {
          id: '2', type: 'v1', label: '2', properties: {},
        },
        {
          id: '3', type: 'v1', label: '3', properties: {},
        },
        {
          id: '4', type: 'v1', label: '4', properties: {},
        },
      ],
      edges: [
        {
          from: '0',
          to: '1',
          type: 'e1',
          weight: 1,
        },
        {
          from: '2',
          to: '1',
          type: 'e1',
          weight: 0.5,
        },
        {
          from: '3',
          to: '1',
          type: 'e1',
          weight: 0.25,
        },
        {
          from: '1',
          to: '4',
          type: 'e1',
          weight: 0.125,
        },
      ],
      // history: [],
    });
  });

  it('can import from JSON', () => {
    const graph2 = new WeightedDirectedGraph('hidden1');
    graph2.fromJSON({
      name: 'hidden1import',
      vertices: [{
        id: '1',
        label: '1',
        type: 'v1',
        properties: {
          color: 'green',
        },
      }, {
        id: '2',
        label: '2',
        type: 'v1',
        properties: {
          color: 'blue',
        },
      }],
      edges: [{
        from: '1',
        to: '2',
        type: 'e1',
        weight: 0.5,
      }],
    });

    graph2.vertices.size.should.equal(2);
    graph2.edgeTypes.size.should.equal(1);
    graph2.edgeTypes.get('e1').size.should.equal(1);

    const v1 = graph2.getVertex('1');
    v1.id.should.equal('1');
    v1.label.should.equal('1');
    v1.type.should.equal('v1');
    v1.properties.size.should.equal(1);
    v1.properties.get('color').should.equal('green');

    const v2 = graph2.getVertex('2');
    v2.id.should.equal('2');
    v2.label.should.equal('2');
    v2.type.should.equal('v1');
    v2.properties.size.should.equal(1);
    v2.properties.get('color').should.equal('blue');

    const e1 = graph2.getEdge('e1', '1', '2');
    v1.edges.from[0].should.equal(e1);
    v2.edges.to[0].should.equal(e1);
  });
});

mochaRun();
